# Copyright (c) Microsoft Corporation. All rights reserved.
"""Classes for collecting summary statistics on the data produced by a Dataflow."""
from collections import namedtuple, OrderedDict
from typing import Any, Dict, List
import json
from .engineapi.api import EngineAPI
from .engineapi.typedefinitions import ActivityReference, DataField, ExecuteInspectorCommonArguments, FieldType, InspectorArguments, ProfileResult, Moments, Quantiles, HistogramBin, ValueCount, TypeCount
from .value import value_from_field

_MAX_ROW_COUNT = 2**31 - 1

ValueCountEntry = namedtuple('ValueCountEntry', ['value', 'count'])
HistogramBucket = namedtuple('HistogramBucket', ['lower_bound', 'upper_bound', 'count'])
TypeCountEntry = namedtuple('TypeCountEntry', ['type', 'count'])

class DTypes(Dict):
    def __init__(self, dict):
        super().__init__(dict)

    def __repr__(self):
        column_types = ["""{0!s:<25}{1!s:<16}""".format(column, dtype, end=' ') for (column, dtype) in self.items()]
        return '\n'.join(column_types)


class DataProfile:
    """
    A DataProfile collects summary statistics on the data produced by a Dataflow.

    :var columns: Profile information for each result column.
    :vartype columns: Dict[str, ColumnProfile]
    """

    def __init__(self):
        self.columns = OrderedDict()

    @classmethod
    def _from_execution(cls, engine_api: EngineAPI, context: ActivityReference):
        table_inspector = engine_api.execute_inspector(ExecuteInspectorCommonArguments(
            context=context,
            inspector_arguments=InspectorArguments(inspector_type='Microsoft.DPrep.TableInspector', arguments={}),
            offset=0,
            row_count=_MAX_ROW_COUNT))
        column_names = [column_definition.id for column_definition in table_inspector.column_definitions]
        rows = table_inspector.rows_data.rows

        FieldAndValue = namedtuple('FieldAndValue', ['field', 'value'])
        def values_for_column(fields):
            return {
                name: FieldAndValue(field=field,
                                    value=value_from_field(field) if field is not None and field.type != FieldType.NULL and field.type != FieldType.ERROR else None)
                for name, field in zip(column_names, fields)
            }

        dp = cls()
        dp.columns = OrderedDict([(row[0].value, ColumnProfile(values_for_column(row))) for row in rows])
        return dp
   
    @classmethod
    def _from_json(cls, json_str):
        dp = cls()
        pr_dict = json.loads(json_str)
        for col in pr_dict:
            dp.columns[col] = ColumnProfile.from_pod(pr_dict[col])
        return dp
        
    @property
    def dtypes(self) -> Dict[str, FieldType]:
        """
        Column data types.

        :return: A dictionary, where key is the column name and value is :class:`azureml.dataprep.FieldType`.
        """
        return DTypes({column: column_profile.type for (column, column_profile) in self.columns.items()})

    @property
    def row_count(self) -> int:
        """
        Count of rows in this DataProfile.

        :return: Count of rows.
        :rtype: int
        """
        for column_profile in self.columns.values():
            # return count of rows from first column profile we iterate over.
            return int(column_profile.count)

    def _to_json(self):
        return json.dumps(self.columns, cls=_DataProfileEncoder)

    def _repr_html_(self):
        """
        HTML representation for IPython.
        """
        try:
            import pandas as pd
        except ImportError:
            return None

        stats = [column_profile.get_stats() for column_profile in self.columns.values()]
        return pd.DataFrame(stats, index=self.columns.keys(), columns=ColumnProfile.STAT_COLUMNS).to_html()

    def __repr__(self):
        return '\n'.join(map(str, self.columns.values()))


class ColumnProfile(ProfileResult):
    """
    A ColumnProfile collects summary statistics on a particular column of data produced by a Dataflow.

    :var column_name: Name of column
    :vartype column_name: str
    :var type: Type of values in column
    :vartype type: FieldType

    :var min: Minimum value
    :vartype min: Any
    :var max: Maximum value
    :vartype max: Any
    :var count: Count of rows
    :vartype count: int
    :var missing_count: Count of rows with a missing value
    :vartype missing_count: int
    :var not_missing_count: Count of rows with a value
    :vartype not_missing_count: int
    :var error_count: Count of rows with an error value
    :vartype error_count: int
    :var percent_missing: Percent of the values that are missing
    :vartype percent_missing: float
    :var empty_count: Count of rows with empty string value
    :vartype empty_count: int

    :var lower_quartile: Estimated 25th-percentile value
    :vartype lower_quartile: float
    :var median: Estimated median value
    :vartype median: float
    :var upper_quartile: Estimated 75th-percentile value
    :vartype upper_quartile: float
    :var mean: Mean
    :vartype mean: float
    :var std: Standard deviation
    :vartype std: float
    :var variance: Variance
    :vartype variance: float
    :var skewness: Skewness
    :vartype skewness: float
    :var kurtosis: Kurtosis
    :vartype kurtosis: float
    :var quantiles: Dictionary of quantiles
    :vartype quantiles: Dict[float, float]

    :var value_counts: Counts of discrete values in the data; None if too many values.
    :vartype value_counts: List[ValueCountEntry]
    :var type_counts: Counts of discrete types in the data.
    :vartype type_counts: List[TypeCountEntry]
    :var histogram: Histogram buckets showing the distribution of the data; None if data is non-numeric.
    :vartype histogram: List[HistogramBucket]
    """
    def __init__(self, values: Dict[str, Any] = None):
        if values is None:
            super(ColumnProfile, self).__init__()
        else:
            field_type = FieldType(values.get('type').value)
            is_numeric = ColumnProfile._is_numeric_type(field_type)
            super(ColumnProfile, self).__init__(
                column_name=values.get('Column').value,
                type=field_type,
                count=values.get('count').value,
                empty_count=values.get('num_empty').value,
                error_count=values.get('num_errors').value,
                missing_count=values.get('num_missing').value,
                not_missing_count=values.get('num_not_missing').value,
                percent_missing=values.get('%missing').value,
                max=values.get('max').field,
                min=values.get('min').field,
                moments=ColumnProfile._prepare_moments(values) if is_numeric else None,
                quantiles=ColumnProfile._prepare_quartiles(values) if is_numeric else None,
                histogram=ColumnProfile._prepare_histogram(values.get('histogram')),
                type_counts=ColumnProfile._prepare_type_counts(values.get('type_count').value),
                value_counts=ColumnProfile._prepare_value_counts(values.get('value_count').value))
        self._quantiles_dict = None

    @property
    def name(self) -> str:
        """
        (Deprecated. Use column_name instead.)
        """
        return self.column_name

    @property
    def kurtosis(self) -> float:
        return None if self.moments is None else self.moments.kurtosis

    @property
    def mean(self) -> float:
        return None if self.moments is None else self.moments.mean
    
    @property
    def skewness(self) -> float:
        return None if self.moments is None else self.moments.skewness
    
    @property
    def std(self) -> float:
        return None if self.moments is None else self.moments.standard_deviation

    @property
    def variance(self) -> float:
        return None if self.moments is None else self.moments.variance

    @property
    def quantiles(self) -> Dict[float, float]:
        if self._quantiles_dict is None:
            quantiles = super(ColumnProfile, self).quantiles
            self._quantiles_dict = OrderedDict([
                (0.001, None if quantiles is None else quantiles.p0_d1),
                (0.01, None if quantiles is None else quantiles.p1),
                (0.05, None if quantiles is None else quantiles.p5),
                (0.25, None if quantiles is None else quantiles.p25),
                (0.50, None if quantiles is None else quantiles.p50),
                (0.75, None if quantiles is None else quantiles.p75),
                (0.95, None if quantiles is None else quantiles.p95),
                (0.99, None if quantiles is None else quantiles.p99),
                (0.999, None if quantiles is None else quantiles.p99_d9)])
        return self._quantiles_dict

    @property
    def lower_quartile(self) -> float:
        return self.quantiles[0.25]

    @property
    def median(self) -> float:
        return self.quantiles[0.50]

    @property
    def upper_quartile(self) -> float:
        return self.quantiles[0.75]

    @property
    def max(self) -> object:
        return value_from_field(DataField.from_pod(super(ColumnProfile, self).max))

    @property
    def min(self) -> object:
        return value_from_field(DataField.from_pod(super(ColumnProfile, self).min))

    @property
    def value_counts(self) -> List[ValueCountEntry]:
        vcs = super(ColumnProfile, self).value_counts
        return None if vcs is None else [ValueCountEntry(value=value_from_field(DataField.from_pod(vc.value)),
                                                         count=vc.count) for vc in vcs]
    
    @property
    def type_counts(self) -> List[TypeCountEntry]:
        tcs = super(ColumnProfile, self).type_counts
        return None if tcs is None else [TypeCountEntry(type=tc.type,
                                                        count=tc.count) for tc in tcs]

    @property
    def histogram(self) -> List[HistogramBucket]:
        bins = super(ColumnProfile, self).histogram
        return None if bins is None else [HistogramBucket(lower_bound=b.lower_bound,
                                                          upper_bound=b.upper_bound,
                                                          count=b.count) for b in bins]

    @property
    def _is_numeric(self) -> bool:
        return ColumnProfile._is_numeric_type(self.type)

    @staticmethod
    def _is_numeric_type(t: FieldType) -> bool:
        return t == FieldType.INTEGER or t == FieldType.DECIMAL

    @staticmethod
    def _prepare_moments(values: Dict[str, Any]):
        return Moments(kurtosis=values.get('kurtosis').value,
                       mean=values.get('mean').value,
                       skewness=values.get('skewness').value,
                       standard_deviation=values.get('standard_deviation').value,
                       variance=values.get('variance').value)

    @staticmethod
    def _prepare_quartiles(values: Dict[str, Any]):
        return Quantiles(p0_d1=values.get('0.1%').value,
                         p1=values.get('1%').value,
                         p5=values.get('5%').value,
                         p25=values.get('25%').value,
                         p50=values.get('50%').value,
                         p75=values.get('75%').value,
                         p95=values.get('95%').value,
                         p99=values.get('99%').value,
                         p99_d9=values.get('99.9%').value)

    @staticmethod
    def _prepare_value_counts(entries: List[Any]):
        if entries and len(entries) > 0:
            return [ValueCount(value=entry['value'], 
                               count=float(entry['count'])) for entry in entries]
        return None

    @staticmethod
    def _prepare_type_counts(entries: List[Any]):
        return [TypeCount(type=FieldType(entry['type']), 
                          count=float(entry['count'])) for entry in entries]

    @staticmethod
    def _prepare_histogram(fields: List[Dict[str, Any]]):
        if fields.value:
            values = (f['value'] for f in fields.value)
            def generate_buckets():
                last_position = None
                for position, count in zip(values, values):
                    if last_position is not None:
                        yield HistogramBin(lower_bound=last_position, upper_bound=position, count=count)
                    last_position = position
            histogram = list(generate_buckets())
            if len(histogram) > 0:
                return histogram
        return None

    def get_stats(self):
        return [
            self.type, self.min, self.max, self.count, self.missing_count, self.not_missing_count, self.percent_missing,
            self.error_count, self.empty_count,
            self.quantiles[0.001] if self._is_numeric else '',
            self.quantiles[0.01] if self._is_numeric else '',
            self.quantiles[0.05] if self._is_numeric else '',
            self.quantiles[0.25] if self._is_numeric else '',
            self.quantiles[0.50] if self._is_numeric else '',
            self.quantiles[0.75] if self._is_numeric else '',
            self.quantiles[0.95] if self._is_numeric else '',
            self.quantiles[0.99] if self._is_numeric else '',
            self.quantiles[0.999] if self._is_numeric else '',
            self.mean if self._is_numeric else '',
            self.std if self._is_numeric else '',
            self.variance if self._is_numeric else '',
            self.skewness if self._is_numeric else '',
            self.kurtosis if self._is_numeric else ''
        ]

    STAT_COLUMNS = [
        'Type', 'Min', 'Max', 'Count', 'Missing Count', 'Not Missing Count', 'Percent missing', 'Error Count',
        'Empty count', '0.1% Quantile', '1% Quantile', '5% Quantile', '25% Quantile', '50% Quantile', '75% Quantile',
        '95% Quantile', '99% Quantile', '99.9% Quantile', 'Mean', 'Standard Deviation', 'Variance', 'Skewness', 'Kurtosis'
    ]

    def _repr_html_(self):
        """
        HTML representation for IPython.
        """
        try:
            import pandas as pd
        except ImportError:
            return None
        return pd.DataFrame(self.get_stats(), index=ColumnProfile.STAT_COLUMNS, columns=['Statistics']).to_html()

    def __repr__(self):
        result = """\
ColumnProfile:
    column_name: {0}
    type: {1}

    min: {2}
    max: {3}
    count: {4}
    missing_count: {5}
    not_missing_count: {6}
    percent_missing: {7}
    error_count: {8}
    empty_count: {9}

""".format(self.column_name,
           self.type,
           self.min,
           self.max,
           self.count,
           self.missing_count,
           self.not_missing_count,
           self.percent_missing,
           self.error_count,
           self.empty_count)

        if self._is_numeric:
            result += """\

    Quantiles:
         0.1%: {0!s}
           1%: {1!s}
           5%: {2!s}
          25%: {3!s}
          50%: {4!s}
          75%: {5!s}
          95%: {6!s}
          99%: {7!s}
        99.9%: {8!s}

    mean: {9!s}
    std: {10!s}
    variance: {11!s}
    skewness: {12!s}
    kurtosis: {13!s}
""".format(self.quantiles[0.001],
           self.quantiles[0.01],
           self.quantiles[0.05],
           self.quantiles[0.25],
           self.quantiles[0.50],
           self.quantiles[0.75],
           self.quantiles[0.95],
           self.quantiles[0.99],
           self.quantiles[0.999],
           self.mean,
           self.std,
           self.variance,
           self.skewness,
           self.kurtosis)

        return result


class _DataProfileEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ColumnProfile):
            return obj.to_pod()
        if isinstance(obj, DataField):
            return obj.to_pod()
        return super(_DataProfileEncoder, self).default(obj)
